import math
import os

import numpy as np

from Constants import PrintOpts, Profiles
from artificial_surf import artificial_surf
from check_environment import print_import_error
from generate_profile import create_2d_profile2, revolve_profile2
from hertz_equations import get_diam_approx
from system_functions import print_it


class ContactBody:
    """Each tribosystems is composed of several contact bodies.
    The ContactBody class is the super class for all types of contact bodies,
    currently rings, balls, disks and flats"""

    # each contact body may use (and reuse) a separate
    # matlab engine for rough surface generation
    matlab_engine = None

    def __init__(self, name, e, ny):
        """All contact bodies have the following attributes"""
        self.name = name
        self.e = e
        self.ny = ny

        self.circle_radius = None
        self.clone = None
        self.contact_width = None
        self.diameter = None
        self.length = None
        self.normal_force = None
        self.path_profile = None
        self.profile = None
        self.rms_roughness = None
        self.rotational_speed = None
        self.roughness_mat = None
        self.true_rms_roughness = None
        self.type_profile = None

        self.res_x = None
        self.delta_x = None
        self.x_axis = None
        self.x_label = None
        self.x_profile = None
        self.file_x_axis = None
        self.file_x_profile = None
        self.r_hertz_x = None

        self.res_y = None
        self.delta_y = None
        self.y_axis = None
        self.y_label = None
        self.y_profile = None
        self.r_hertz_y = None

        self.res_z = None
        self.delta_z = None
        self.z_axis = None
        self.z_label = None
        self.z_profile = None

    def make_slave(self, contact_body):
        """Changes the contact body (slave) grid size and resolution to match
        that of another contact body (master)"""
        self.length = contact_body.length
        self.contact_width = contact_body.contact_width

        self.res_x = contact_body.res_x
        self.delta_x = contact_body.delta_x
        self.x_axis = contact_body.x_axis
        self.x_label = contact_body.x_label

        self.res_y = contact_body.res_y
        self.delta_y = contact_body.delta_y
        self.y_axis = contact_body.y_axis
        self.y_label = contact_body.y_label

        self.res_z = contact_body.res_z
        self.delta_z = contact_body.delta_z
        self.z_axis = contact_body.z_axis
        self.z_label = contact_body.z_label

    def get_hertz_radii(self):
        """Get an approximation of the hertz contact radii from the true
        contact body geometry"""
        self.r_hertz_x = get_diam_approx(self.x_axis, self.x_profile) / 2
        self.r_hertz_y = get_diam_approx(self.y_axis, self.y_profile) / 2

    def get_profile(self):
        """create a 2D profile, then revolve the 2D profile to obtain 3D
        profile"""
        self.x_axis, self.x_profile, self.file_x_axis, self.file_x_profile, self.delta_x = \
            create_2d_profile2(self.diameter, self.length, self.type_profile,
                               self.x_axis, self.res_x,
                               self.path_profile, self.circle_radius)
        self.profile, self.y_profile = \
            revolve_profile2(self.diameter, self.x_profile, self.y_axis,
                             self.res_x, self.res_y)

    def get_area(self, pressure_field):
        """Calculate true contact area based on number of non-zero pressure
        elements"""
        return len(
            pressure_field[pressure_field > 0]) * self.delta_x * self.delta_y

    def add_rms_roughness(self):
        """Add roughness to the contact body geometry. Currently, the roughness
        distribution is generated in matlab, hence, a matlab engine is required.
        A python implementation is on the TODO list"""
        if self.matlab_engine is None:
            try:
                print_it(
                    "trying to start matlab engine for {} "
                    "rough surface generation; this might take a while".
                    format(self.name), PrintOpts.lvl1.value)
                import matlab.engine
                self.matlab_engine = matlab.engine.start_matlab()
                self.matlab_engine.cd(os.getcwd(), nargout=0)
            except (ImportError, OSError):
                print_import_error('matlab engine for python',
                                   PrintOpts.lvl1.value)
                print_it(
                    "skipping roughness generation due to "
                    "missing matlab engine",
                    PrintOpts.lvl1.value)
                self.matlab_engine = 'no engine found'
        elif self.matlab_engine != 'no engine found':
            sigma = self.rms_roughness / 1000 / 1000
            h = 0.8
            m = self.res_y + 1
            lx = (self.contact_width + self.contact_width * self.res_y / (
            self.res_y + 1)) / 1000
            n = round(self.length / self.contact_width * m)
            qr = (2 * math.pi) / (self.contact_width / 20)
            z, pixel_width, qx, qy, cq, q, c, n, m = \
                artificial_surf(sigma, h, lx, m, n, qr, self.matlab_engine)

            extract_n = np.round(np.linspace(0, n - 1, self.res_x)).astype(int)
            self.roughness_mat = z[extract_n, 0:-1]
            self.true_rms_roughness = np.sqrt(
                np.mean(np.square(self.roughness_mat)))
            self.profile = np.add(self.profile, self.roughness_mat)
        else:
            pass


class Flat(ContactBody):
    """Contact body with completely flat surface and no specific physical size.
    This contact body has no 'make_profile' method since it is only supposed to
    be used as a slave to a Ball or Ring type contact body"""

    def __init__(self, name, e, ny, rms_rough=None):
        super().__init__(name, e, ny)
        self.type_profile = Profiles.NoProfile.value
        self.rms_roughness = rms_rough

    def make_slave_to(self, contact_body):
        """function call to 'make_slave' method of super class, plus some
        Flat-specific adjustments"""
        self.make_slave(contact_body)
        self.x_profile = np.zeros(self.res_x)
        self.y_profile = np.zeros(self.res_y)
        self.profile = np.zeros((self.res_x, self.res_y))
        self.get_hertz_radii()
        if self.rms_roughness is not None:
            self.add_rms_roughness()

    def simple_clone(self):
        """make a simple clone of contact body that is used for grid size
        determination"""
        self.clone = Flat('simple clone', self.e, self.ny)


class Ball(ContactBody):
    """Contact body with ball geometry"""

    def __init__(self, name, e, ny, diameter, rms_rough=None):
        super().__init__(name, e, ny)
        self.diameter = diameter
        self.type_profile = Profiles.Circle.value
        self.circle_radius = self.diameter / 2
        self.length = diameter
        self.r_hertz_y = diameter
        self.rms_roughness = rms_rough

    def make_profile(self, res_x, res_y, normal_force, contact_width=None):
        """mainly function call to 'get_profile' method of super class, plus
        some Ball-specific adjustments"""
        self.res_x = res_x
        self.res_y = res_y
        self.x_label = 'length in mm'
        self.y_label = 'width in mm'
        self.z_label = 'profile in mm'
        self.contact_width = \
            contact_width or get_contact_width(self.diameter, normal_force,
                                               self.ny, self.e, self.length)
        self.x_axis, self.delta_x = np.linspace(-self.contact_width / 2,
                                                self.contact_width / 2, res_x,
                                                retstep=True)
        self.y_axis, self.delta_y = np.linspace(-self.contact_width / 2,
                                                self.contact_width / 2, res_y,
                                                retstep=True)

        self.get_profile()
        self.get_hertz_radii()
        if self.rms_roughness is not None:
            self.add_rms_roughness()

    def make_slave_to(self, contact_body):
        """function call to 'make_slave' method of super class, plus some
        Ball-specific adjustments"""
        self.make_slave(contact_body)
        self.get_profile()
        self.get_hertz_radii()
        if self.rms_roughness is not None:
            self.add_rms_roughness()

    def simple_clone(self):
        """make a simple clone of contact body that is used for grid size
        determination"""
        self.clone = Ball('simple clone', self.e, self.ny, self.diameter)


class Disk(ContactBody):
    """Contact body with disk geometry. Difference to Flat-type contact body is
    that the physical size is explicitly
    defined here"""

    def __init__(self, name, e, ny, diameter=0, length=None, rot_vel=0,
                 rms_rough=None):
        super().__init__(name, e, ny)
        self.diameter = diameter
        self.rot_vel = rot_vel
        self.rms_roughness = rms_rough
        self.length = length

    def make_profile(self, res_x, res_y, normal_force, contact_width=None):
        """mainly function call to 'get_profile' method of super class, plus
        some Disk-specific adjustments"""
        self.res_x = res_x
        self.res_y = res_y
        self.x_label = 'length in mm'
        self.y_label = 'width in mm'
        self.z_label = 'profile in mm'
        self.contact_width = \
            contact_width or get_contact_width(self.diameter, normal_force,
                                               self.ny, self.e, self.length)
        self.x_axis, self.delta_x = np.linspace(-self.contact_width / 2,
                                                self.contact_width / 2, res_x,
                                                retstep=True)
        self.y_axis, self.delta_y = np.linspace(-self.contact_width / 2,
                                                self.contact_width / 2, res_y,
                                                retstep=True)

        self.x_profile = np.zeros(self.res_x)
        self.y_profile = np.zeros(self.res_y)
        self.profile = np.zeros((self.res_x, self.res_y))

        self.get_hertz_radii()
        if self.rms_roughness is not None:
            self.add_rms_roughness()

    def make_slave_to(self, contact_body):
        """function call to 'make_slave' method of super class, plus some
        Disk-specific adjustments"""
        self.make_slave(contact_body)
        self.x_profile = np.zeros(self.res_x)
        self.y_profile = np.zeros(self.res_y)
        self.profile = np.zeros((self.res_x, self.res_y))
        self.get_hertz_radii()
        if self.rms_roughness is not None:
            self.add_rms_roughness()

    def simple_clone(self):
        """make a simple clone of contact body that is used for grid size
        determination"""
        self.clone = Disk('simple clone', self.e, self.ny)


class Ring(ContactBody):
    """Ring type contact body. Can be used to generate ring, roller and
    arbitrarily-shaped contact bodies"""

    def __init__(self, name, e, ny, diameter, length, type_profile,
                 rms_rough=None, circle_radius=None,
                 path_profile=None):
        super().__init__(name, e, ny)
        self.type_profile = type_profile
        self.diameter = diameter
        self.length = length
        self.r_hertz_y = diameter
        self.psi = 0
        self.profile = None
        self.rms_roughness = rms_rough
        self.circle_radius = circle_radius
        self.path_profile = path_profile

    def make_profile(self, res_x, res_y, normal_force, contact_width=None,
                     contact_length=None):
        """mainly function call to 'get_profile' method of super class, plus
        some Ring-specific adjustments"""
        self.res_x = res_x
        self.res_y = res_y
        self.x_label = 'length in mm'
        self.y_label = 'width in mm'
        self.z_label = 'profile in mm'
        self.contact_width = \
            contact_width or get_contact_width(self.diameter, normal_force,
                                               self.ny, self.e, self.length)
        if contact_length is None:
            contact_length = self.length
        self.x_axis, self.delta_x = np.linspace(-contact_length / 2,
                                                contact_length / 2, res_x,
                                                retstep=True)
        self.y_axis, self.delta_y = np.linspace(-self.contact_width / 2,
                                                self.contact_width / 2, res_y,
                                                retstep=True)

        self.get_profile()
        self.get_hertz_radii()
        if self.rms_roughness is not None:
            self.add_rms_roughness()

    def make_slave_to(self, contact_body):
        """function call to 'make_slave' method of super class, plus some
        Ring-specific adjustments"""
        self.make_slave(contact_body)
        self.get_profile()
        self.get_hertz_radii()
        if self.rms_roughness is not None:
            self.add_rms_roughness()

    def simple_clone(self):
        """make a simple clone of contact body that is used for grid size
        determination"""
        self.clone = Ring('simple clone', self.e, self.ny, self.diameter,
                          self.length, Profiles.ISO.value)


def get_contact_width(diameter, normal_force, ny, e, length):
    """Very rough hertz contact width calculation for initial guess of grid
    size"""
    r_hertz = diameter / 2
    b = math.sqrt(
        1.5 * normal_force * r_hertz * (1 - math.pow(ny, 2)) / (2 * e * length))
    return 2 * b



# -*- coding: utf-8 -*-

"""

This module contains functions related to Hertz contact theory.

"""

from math import sqrt, pi, log, floor


def __rred(r_1, r_2):
    """

    Calculate the reduced (effective) radius of two radii according to Hertzian
    contact theory.

    Parameters
    ----------
    r_1: scalar
        The first radius.
    r_2: scalar
        The second radius.

    Returns
    -------
    r_red: scalar
        The reduced (effective) radius.

    """
    if (r_1 == 0 or abs(r_1) == float('inf')) and r_2 != 0:
        r_red = r_2
    elif (r_2 == 0 or abs(r_2) == float('inf')) and r_1 != 0:
        r_red = r_1
    elif (r_1 == 0 or abs(r_1) == float('inf')) and \
            (r_2 == 0 or abs(r_2) == float('inf')):
        r_red = 0
    elif r_1 == -r_2:
        r_red = 0
    else:
        r_red = 1 / (1 / r_1 + 1 / r_2)
    return r_red


def reff(r_x_1, r_y_1, r_x_2, r_y_2, ):
    """

    Calculate the effective radii for two bodies according to Hertzian contact
    theory. It is assumed that the two major axis of each body (x- and y-axis)
    are perpendicular to each other and that the x and y axes of both bodies are
    aligned.

    Parameters
    ----------
    r_x_1: scalar
        The radius of body 1 in direction 1 (x).
    r_y_1: scalar
        The radius of body 1 in direction 2 (y).
    r_x_2: scalar
        The radius of body 2 in direction 1 (x).
    r_y_2: scalar
        The radius of body 2 in direction 2 (y).

    Returns
    -------
    r_eff: scalar
        The effective radius.
    r_eff_x: scalar
        The effective radius in x-direction.
    r_eff_y: scalar
        The effective radius in y-direction.

    """
    recip_radius = []
    for radius in [r_x_1, r_y_1, r_x_2, r_y_2]:
        if radius == 0:
            recip_radius.append(float('inf'))
        else:
            recip_radius.append(1 / radius)

    r_eff_x = __rred(r_x_1, r_x_2)
    r_eff_y = __rred(r_y_1, r_y_2)
    r_eff = __rred(r_eff_x, r_eff_y)
    return r_eff, r_eff_x, r_eff_y


def eeff(e_1, nu_1, e_2, nu_2):
    """

    Calculate the effective (Young's) modulus of two contact bodies according
    to Hertzian contact theory.

    Parameters
    ----------
    e_1: ndarray, scalar
        The Young's modulus of contact body 1.
    nu_1: ndarray, scalar
        The Poisson ratio of contact body 1.
    e_2: ndarray, scalar
        The Young's modulus of contact body 2.
    nu_2: ndarray, scalar
        The Poisson ratio of contact body 1.

    Returns
    -------
    e_eff: scalar
        The effective modulus.

    """
    e_eff = 1 / ((1 - nu_1 ** 2) / (2 * e_1) + (1 - nu_2 ** 2) / (2 * e_2))
    return e_eff


def __auxparamshertz(r_eff_x, r_eff_y):
    """

    Calculate a set of parameters required for Hertz contact area/pressure
    calculations.

    Parameters
    ----------
    r_eff_x: scalar
        The effective redius of the contact problem in x-direction.
    r_eff_y: scalar
        The effective redius of the contact problem in x-direction.

    Returns
    -------
    a_ast: scalar
        A parameter often referred to as a*.
    b_ast: scalar
        A parameter often referred to as b*.
    kappa: scalar
        A parameter often referred to as kappa.
    param_lambda: scalar
        A parameter often referred to as lambda.

    """
    if r_eff_x == 0 or r_eff_y == 0:
        param_lambda = 0
    else:
        param_lambda = min(r_eff_x / r_eff_y, r_eff_y / r_eff_x)
    kappa = 1 / (1 + sqrt(log(16 / param_lambda) / (2 * param_lambda)) -
                 sqrt(log(4)) + 0.16 * log(param_lambda))
    a_ast = kappa * (1 + (2 * (1 - kappa ** 2)) / (pi * kappa ** 2) -
                     0.25 * log(kappa)) ** (1 / 3)
    b_ast = a_ast / kappa
    return a_ast, b_ast, kappa, param_lambda


def dhertz(e_eff, r_x_1, r_y_1, r_x_2, r_y_2, force):
    """

    Calculate the elastic normal displacement of a contact problem according
    to Hertzian contact theory.

    Parameters
    ----------
    e_eff: scalar
        The effective modulus of the contact problem.
    r_x_1: scalar
        The radius of body 1 in direction 1 (x).
    r_y_1: scalar
        The radius of body 1 in direction 2 (y).
    r_x_2: scalar
        The radius of body 2 in direction 1 (x).
    r_y_2: scalar
        The radius of body 2 in direction 2 (y).
    force: scalar
        The normal force in the contact.

    Returns
    -------
    norm_disp: scalar
        The elastic normal displacement of the contact problem.

    """
    apb = 0.5 * (1 / r_x_1 + 1 / r_y_1 + 1 / r_x_2 + 1 / r_y_2)
    bma = 0.5 * sqrt((1 / r_x_1 - 1 / r_y_1) ** 2 +
                     (1 / r_x_2 - 1 / r_y_2) ** 2)
    r_a = 1 / (apb - bma)
    r_b = 1 / (apb + bma)
    r_c = sqrt(r_a * r_b)

    f_1 = 1 - pow(pow(r_a / r_b, 0.0602) - 1, 1.456)
    f_2 = 1 - pow(pow(r_a / r_b, 0.0684) - 1, 1.531)

    param_c = pow(3 * force * r_c / (4 * e_eff), 1 / 3) * f_1
    param_e = 1 - pow(r_b / r_a, 4 / 3)
    param_a = param_c * pow(1 - param_e ** 2, 1 / 4)
    param_b = param_c * pow(1 - param_e ** 2, 1 / 4)
    norm_disp = param_a * param_b / r_c * (f_2 / f_1)
    return norm_disp


def ahertz(r_eff, r_eff_x, r_eff_y, e_eff, force):
    """

    Calculate the contact area according to Hertzian contact theory.

    Parameters
    ----------
    r_eff: scalar
        The effective radius of the contact problem.
    r_eff_x: scalar
        The effective radius of the contact problem in x-direction.
    r_eff_y: scalar
        The effective radius of the contact problem in y-direction.
    e_eff: scalar
        The effective modulus of the contact problem.
    force: scalar
        The normal force in the contact.

    Returns
    -------
    half_axis_a: scalar
        The contact area half axis in x-direction.
    half_axis_b: scalar
        The contact area half axis in y-direction.
    a_hertz: scalar
        The contact area.

    """
    a_ast, b_ast, _, _ = __auxparamshertz(r_eff_x, r_eff_y)
    half_axis_a = a_ast * (3 * force * r_eff / e_eff) ** (1 / 3)
    half_axis_b = b_ast * (3 * force * r_eff / e_eff) ** (1 / 3)
    a_hertz = pi * half_axis_a * half_axis_b
    return half_axis_a, half_axis_b, a_hertz


def fhertz(r_eff, r_eff_x, r_eff_y, e_eff, p_critical):
    """

    Calculate the load carrying capacity of a Hertzian contact.

    Parameters
    ----------
    r_eff: scalar
        The effective radius of the contact problem.
    r_eff_x: scalar
        The effective radius of the contact problem in x-direction.
    r_eff_y: scalar
        The effective radius of the contact problem in y-direction.
    e_eff: scalar
        The effective modulus of the contact problem.
    p_critical: scalar
        The critical mean Hertzian contact pressure that the contact can sustain
        without plastic deformation.

    Returns
    -------
    f_crit: scalar
        The load carrying capacity of the Hertzian contact.

    """
    a_ast, b_ast, _, _ = __auxparamshertz(r_eff_x, r_eff_y)
    f_crit = (pi * a_ast * b_ast * p_critical) ** 3 * (3 * r_eff / e_eff) ** 2
    return f_crit


def phertz(r_eff, r_eff_x, r_eff_y, e_eff, force):
    """

    Calculate the mean contact pressure in a Hertzian contact.

    Parameters
    ----------
    r_eff: scalar
        The effective radius of the contact problem.
    r_eff_x: scalar
        The effective radius of the contact problem in x-direction.
    r_eff_y: scalar
        The effective radius of the contact problem in y-direction.
    e_eff: scalar
        The effective modulus of the contact problem.
    force: scalar
        The normal force in the contact.

    Returns
    -------
    p_hertz: scalar
        The mean Hertzian contact pressure in the contact.

    """
    _, _, area = ahertz(r_eff, r_eff_x, r_eff_y, e_eff, force)
    p_hertz = force / area
    return p_hertz


def approx_hertz_rad(axis, profile):
    """

    Approximate the Hertz contact radius of a profile, i.e., find the radius
    of a circle that minimizes the average discrepancy between the actual
    profile and a circle profile.

    Parameters
    ----------
    axis: ndarray
        The coordinates of the `profile` points along their reference axis. For
        example, if the `profile` array contains the profile heights along the
        x-axis in the range [-5, 5], then `axis` contains the corresponding
        x-axis values.
    profile: ndarray
        The profile heights along `axis`.

    Returns
    -------
    rad: scalar or inf
        The radius of the circle that best approximates `profile`. If the
        profile cannot be approximated with a circle (usually of the profile
        is a straight line), rad is equal to `inf`.

    """
    x_1 = axis[0]
    x_2 = axis[floor(len(axis) / 2)]
    x_3 = axis[-1]
    y_1 = profile[0]
    y_2 = profile[floor(len(axis) / 2)]
    y_3 = profile[-1]

    a = sqrt((x_1 - x_2) ** 2 + (y_1 - y_2) ** 2)
    b = sqrt((x_2 - x_3) ** 2 + (y_2 - y_3) ** 2)
    c = sqrt((x_3 - x_1) ** 2 + (y_3 - y_1) ** 2)
    s = (a + b + c) / 2
    d = sqrt(s * (s - a) * (s - b) * (s - c))

    try:
        rad = a * b * c / (4 * d)
    except ZeroDivisionError:
        rad = float('Inf')
    return rad



# -*- coding: utf-8 -*-
"""

Functions related to roller bearings.

"""

from math import tan, pi

import numpy as np

from tribology.boundary_element import __secant


def fcylrolbear(ang_pos, comb_prof, ax_rol, f_rad, rad_clear=0, max_dif=0.0005):
    """

    Calculate the normal force distribution in a cylindrical roller bearing
    according to DIN 26281. For the calculation, each roller is modelled as a
    set of neighbouring spring slices with known stiffness. Applying an outer
    force :code:`f_rad` leads to an elastic displacement of the spring slices
    of each roller. To find the normal force on each roller, the spring slice
    displacement is varied systematically until the sum of the inner spring
    forces is (approximately) equal to the outer force :code:`f_rad`.

    Parameters
    ----------
    ang_pos: ndarray
        The angular positions of the rollers in angular coordinates. The length
        of the array determines the number of rollers.
    comb_prof: ndarray
        The combined 1D profile of the roller and the bearing inner/outer ring.
    ax_rol: ndarray
        The longitudinal axis of the roller with length :code:`len(comb_prof)`.
        The length of the axis determines the number of roller spring slices.
    f_rad: scalar
        The normal force acting on the inner ring of the bearing. The force acts
        from the center of the bearing in the direction of the origin of the
        polar axis.
    rad_clear: scalar, optional
        The radial clearance of the bearing. Default is 0.
    max_dif: scalar, optional
        The maximum allowed difference between the outer force :code:`f_rad` and
        the sum over all roller forces (in percent of :code:`f_rad`). Default
        is 0.05 %.

    Returns
    -------
    f_rols: ndarray
        The roller normal forces for each roller position in :code:`ang_pos`.
    disp_rols: ndarray
        The elastic normal displacement in the ring-roller contact for each
        roller position in :code:`ang_pos`.
    delta_f: scalar
        The difference between the outer force :code:`f_rad` and the sum over
        all roller forces in the direction of applied force, in units of force
        (i.e., the difference between user input force and the force used for
        the numerical calculation).

    """

    if f_rad == 0:
        return np.zeros(len(ang_pos)), np.zeros(len(ang_pos)), 0

    x_axis = abs(ax_rol)
    len_roll = max(ax_rol) - min(ax_rol)
    res_x = len(x_axis)
    num_rols = len(ang_pos)
    cos_pos = np.cos(ang_pos)

    # stiffness constants according to DIN 26281
    stiff_rol = 35948 * len_roll**(8 / 9)
    stiff_slice = stiff_rol / res_x

    sum_re_ns = np.zeros(num_rols)
    psi_j = np.zeros(num_rols)
    delta_j = np.zeros(num_rols)
    sum_mns = np.zeros(num_rols)
    delta_jk = np.zeros((num_rols, res_x))
    disp_rols = np.zeros(num_rols)

    sec = [[], []]

    delta_r = rad_clear + rad_clear * 10**-6
    delta_f = 20

    while abs(delta_f) > max_dif * f_rad:
        sum_slice_disps = 0
        disp_rols = np.zeros(num_rols)

        for rol in range(num_rols):
            sum_re_ns[rol] = 0
            sum_mns[rol] = 0
            delta_j[rol] = delta_r * cos_pos[rol] - rad_clear / 2

            for idx in range(res_x):
                delta_jk[rol, idx] = max((delta_j[rol] - x_axis[idx] *
                                          tan(psi_j[rol]) -
                                          2 * comb_prof[idx]), 0)
                sum_re_ns[rol] += pow(delta_jk[rol, idx], (10 / 9))

            disp_rols[rol] = cos_pos[rol] * sum_re_ns[rol]
            sum_slice_disps += disp_rols[rol]

        delta_f = f_rad - stiff_slice * sum_slice_disps
        sec[0] = np.append(sec[0], [delta_r])
        sec[1] = np.append(sec[1], [delta_f])
        delta_r = __secant(sec[0], sec[1])

    f_rols = sum_re_ns * stiff_slice

    return f_rols, disp_rols, delta_f


def kinaxthrustrolbear(rot_vel_1, mean_diam, ax_rol, diam_rol, rot_vel_2=0):
    """

    Calculate the kinematics of an axial thrust roller bearing, for example
    a bearing of type 81212. This function calculates the effective velocity
    along the disc (washer) and roller raceway as well as the slip (in percent)
    along the roller's axis of rotation.

    Parameters
    ----------
    rot_vel_1: scalar
        The rotational velocity (in rpm) of the shaft washer of the bearing.
    mean_diam: positive scalar
        The mean diameter of the bearing (the diameter of the bearing where
        there is zero slip in the contact).
    ax_rol: ndarray
        The axis of rotation of the roller, typically spanning from -`length`/2
        to +`length`/2, where `length` is the length of the roller (or the
        section of the roller length that is of interest for the calculation,
        i.e., the points can be asymmetric about the center of the roller).
    diam_rol: positive scalar
        The diameter of the roller.
    rot_vel_2: scalar
        The rotational velocity (in rpm) of the housing washer of the bearing.
        By default, it is assumed that the housing washer is not rotating, so
        `rot_vel_2 = 0`

    Returns
    -------
    raceway_vel_eff: ndarray
        The effective velocity of the shaft washer along the roller axis.
    raceway_vel_rol: ndarray
        The raceway velocity of the roller along the roller axis.
    slip_rol: ndarray
        The roller slip (in percent) along the roller axis. This parameter is
        calculated by comparing the raceway velocity of the roller (at each
        point along its axis) to the ideal rolling velocity in the same point
        (the velocity that would lead to zero slip).

    """
    omega = (rot_vel_1 - rot_vel_2) / 60
    omega_cage = omega / 2
    omega_rol = mean_diam / diam_rol * omega_cage
    raceway_vel_eff = 2 * pi * (mean_diam / 2 + ax_rol) * omega
    raceway_vel_rol = 2 * pi * (mean_diam / 2 + ax_rol) * omega_cage \
        + pi * diam_rol * omega_rol
    rel_vel = pi * diam_rol * omega_rol - 2 * pi * \
        (mean_diam / 2 + ax_rol) * omega_cage
    slip_rol = rel_vel / (2 * pi * (mean_diam / 2 + ax_rol) * omega_cage)
    return raceway_vel_eff, raceway_vel_rol, slip_rol



import math
import os
import sys

import scipy.io

from Constants import PrintOpts, SubDir


def in_d(key, dic, return_if_false=None):
    """Checks if key is in dic. returns corresponding value if key is found;
    returns None (default) or return_if_false (if specified) otherwise"""
    if key in dic:
        return dic[key]
    else:
        if return_if_false is None:
            return None
        else:
            return return_if_false


def exit_program(reason):
    """Exits program and raises exception with text 'reason'"""
    print("\np3can error:\n")
    raise SystemExit(reason)


def print_it(message, level=PrintOpts.lvl0.value):
    """Pre-pends string 'level' to message and prints combined string to
    terminal"""
    if os.environ["AUTO_PRINT"] == "True":
        print("{}{}".format(level, message))


def make_directory(base_dir, directory_name):
    """Makes sub-directory of base_dir with name directory_name. Checks if
    directory exists already, then creates it if
    necessary. Returns path of new directory."""
    dir_to_create = os.sep.join([base_dir, directory_name])
    if os.path.isdir(dir_to_create) is False:
        os.mkdir(dir_to_create)
    return dir_to_create


def print_progress(current, to_reach):
    """Print progress bar to terminal based on ratio of (current:to_reach)"""
    current += 1
    prog_bar_len = 25
    approximate_progress = int(math.floor(current / to_reach * prog_bar_len))
    progress_bar = approximate_progress * '#' + '' + \
                   (prog_bar_len - approximate_progress) * ' '
    if os.environ["AUTO_PRINT"] == "True":
        sys.stdout.write(
            "\r{}progress: |{}| ({} %){}".format(PrintOpts.lvl1.value,
                                                 progress_bar,
                                                 round(
                                                     current / to_reach * 100),
                                                 '' if current != to_reach
                                                 else '\n'))
        sys.stdout.flush()
    return current


def print_rel_prox_to_sol(resulting_force, normal_load, threshold_factor):
    """Print progress bar to indicate proximity of half space solution to
    acceptable solution"""
    prog_bar_len = 25
    rel_dist = abs((resulting_force - normal_load) / normal_load)
    to_reach = 1
    proximity = 1 - rel_dist
    try:
        approximate_progress = int(round(proximity / to_reach * prog_bar_len))
    except ValueError:
        approximate_progress = 1
    progress_bar = approximate_progress * '#' + '' + min(
        (prog_bar_len - approximate_progress), prog_bar_len - 1) * ' '
    if abs((resulting_force - normal_load) / normal_load) < abs(
            threshold_factor):
        append_str = '\n'
    else:
        append_str = ''
    if os.environ["AUTO_PRINT"] == "True":
        sys.stdout.write(
            "\r{}progress: |{}| ({} N away from solution){}".format(
                PrintOpts.lvl1.value, progress_bar,
                round(abs((resulting_force - normal_load)), 1), append_str))
        sys.stdout.flush()
    return proximity


def get_p3can_version(plot_level=PrintOpts.lvl0.value, raise_dir=0):
    """Reads program version number from build.txt file and returns it. Prints
    warning message to terminal if build.txt
    not found"""
    try:
        with open(('..{}'.format(os.sep)) * raise_dir + '..{}build.txt'.format(
                os.sep)) as build_file:
            return build_file.readline()
    except FileNotFoundError:
        print_it(
            "warning: could not find file 'build.txt'. continue running unknown"
            " p3can version.",
            plot_level)
        return "<unknown>"


def save_to_matlab(dat_dict, directory, file_name):
    """Saves all variables in dat_dict to one common matlab database with name
    file_name. the variable names in the matlab database correspond to the key
    in dat_dict; the variable values corresponds to the value in dat_dict"""
    scipy.io.savemat(os.sep.join(
        [make_directory(directory, SubDir.matlab.value), file_name]) + '.mat',
                     dat_dict)


def to_preci(x, p):
    """
    returns a string representation of x formatted with a precision of p

    Originally based on the webkit javascript implementation taken from here:
    https://code.google.com/p/webkit-mirror/source/browse/
                            JavaScriptCore/kjs/number_object.cpp

    I found the below code here:
    http://randlet.com/blog/python-significant-figures-format/

    """

    x = float(x)

    if x == 0.:
        return "0." + "0" * (p - 1)

    out = []

    if x < 0:
        out.append("-")
        x = -x

    e = int(math.log10(x))
    tens = math.pow(10, e - p + 1)
    n = math.floor(x / tens)

    if n < math.pow(10, p - 1):
        e = e - 1
        tens = math.pow(10, e - p + 1)
        n = math.floor(x / tens)

    if abs((n + 1.) * tens - x) <= abs(n * tens - x):
        n = n + 1

    if n >= math.pow(10, p):
        n = n / 10.
        e = e + 1

    m = "%.*g" % (p, n)

    if e < -2 or e >= p:
        out.append(m[0])
        if p > 1:
            out.append(".")
            out.extend(m[1:p])
        out.append('e')
        if e > 0:
            out.append("+")
        out.append(str(e))
    elif e == (p - 1):
        out.append(m)
    elif e >= 0:
        out.append(m[:e + 1])
        if e + 1 < len(m):
            out.append(".")
            out.extend(m[e + 1:])
    else:
        out.append("0.")
        out.extend(["0"] * -(e + 1))
        out.append(m)

    return "".join(out)
